import {$on} from '../shared'
import {useSocket} from './socket'

let apiUrl: string | undefined;
let accessToken: string | undefined;
let artifact: string | undefined;

export async function configApi(cfg: Partial<ApiConfig>) {
  if (
    cfg.accessToken !== accessToken ||
    cfg.apiUrl !== apiUrl ||
    cfg.artifact !== artifact
  ) {
    await ajax(cfg.apiUrl + "/" + cfg.artifact, {
      headers: {"Authorization": `Bearer ${cfg.accessToken}`}
    })

    accessToken = cfg.accessToken;
    artifact = cfg.artifact;
    apiUrl = cfg.apiUrl;

    useSocket(cfg)
  }
}

export async function ajax<T>(url: string, init: RequestInit = {}): Promise<T> {
  return new Promise<T>(async (resolve, reject) => {
    if (accessToken) {
      const headers = new Headers(init.headers);
      headers.set("Authorization", `Bearer ${accessToken}`);
      init.headers = headers;
    }
    try {
      const uri = (apiUrl ?? '') + url;
      const response = await fetch(uri, init);
      const res = await response.json();
      if (
        res != null &&
        typeof res === "object" &&
        !Array.isArray(res) &&
        "ok" in res &&
        "msg" in res &&
        typeof res.ok === "boolean" &&
        typeof res.msg === "string"
      ) {
        if (res.ok) {
          resolve(res.data);
        } else {
          reject(new Error(res.msg));
        }
      } else {
        reject(new Error("服务器异常"));
      }
    } catch (err) {
      // 服务器返回的有可能不是 JSON 数据
      if (err instanceof SyntaxError) {
        reject(new Error("服务器异常"));
      } else if (err instanceof Error && err.message === "Failed to fetch") {
        reject(new Error("服务器异常"));
      } else {
        reject(err);
      }
    }
  });
}

export interface UploadArgs {
  skip: false;
  token: string;
  key: string;
}

export interface UploadSkip {
  skip: true;
}

/**
 * 初始化任务
 */
export function initiateUploadApi(
  fileMime: string,
  fileHash: string,
  signal: AbortSignal,
): Promise<UploadSkip | UploadArgs> {
  const url = `/${artifact}/objects/initiate`;
  const body = new URLSearchParams({fileMime, fileHash});
  const init = {method: "POST", body, signal};
  return ajax<UploadSkip | UploadArgs>(url, init);
}

export function completeUploadApi(
  fileHash: string,
  fileNames: string[],
  directories: number[],
  signal: AbortSignal,
) {
  const targets: Map<number, Set<string>> = new Map();
  for (let i = 0; i < directories.length; i++) {
    const dirid = directories[i];
    const dirs = targets.get(dirid) ?? new Set();
    if (!fileNames[i]) return Promise.reject("参数错误");
    dirs.add(fileNames[i]);
    targets.set(dirid, dirs);
  }
  const body = new URLSearchParams({fileHash});
  targets.forEach((names, dir) => {
    names.forEach((name) => {
      body.append("directoryId", `${dir}`);
      body.append("fileName", name);
    });
  });
  const url = `/${artifact}/objects/complete`;
  return ajax<
    Array<{
      id: TaskId;
      directoryId: number;
      fileName: string;
      fileHash: string;
      createdAt: string;
      updatedAt: string;
    }>
  >(url, {method: "POST", body, signal});
}

interface SimulateQiniuCallbackOptions {
  key: string;
  name: string;
  mime: string;
  md5: FileHash;
  size: number;
}

export function simulateQiniuCallback(
  data: SimulateQiniuCallbackOptions,
  signal: AbortSignal,
) {
  const url = `/${artifact}/objects/uploaded`;
  const body = JSON.stringify({...data, etag: "<etag>", bucket: "<bucket>"});
  return ajax<any>(url, {
    method: "POST",
    body,
    signal,
    headers: {"Content-Type": "application/json"},
  });
}

// 新建文件夹
$on('dir', 'create', (data: NewDirEventData): Promise<CloudDirectory> => {
  const {title, pid = 0} = data
  const uri = `/${artifact}/directories`;
  const body = new URLSearchParams({title});
  if (pid > 0) body.set("parentId", `${pid}`);
  return ajax<CloudDirectory>(uri, {method: "POST", body});
})

// 所有文件夹
$on('dir', 'all', (): Promise<CloudDirectory[]> => {
  return ajax<CloudDirectory[]>(`/${artifact}/directories`);
})

// 删除文件夹
$on('dir', 'delete', (id: number): Promise<boolean> => {
  const url = `/${artifact}/directories/${id}`;
  return ajax<boolean>(url, {method: "DELETE"});
})

// 重命名文件夹
$on('dir', 'rename', (data: RenameDirEventData): Promise<CloudDirectory> => {
  const url = `/${artifact}/directories/${data.id}`;
  const body = new URLSearchParams({title: data.title});
  return ajax<CloudDirectory>(url, {method: "PUT", body});
})

// 文件重命名
$on('file', 'rename', async (data: RenameFileEventData): Promise<CloudFile> => {
  const url = `/${artifact}/files/${data.id}`;
  const body = new URLSearchParams({name: data.name});
  return await ajax(url, {method: "PUT", body});
})

// 删除文件
$on('file', 'delete', (files: number[] | number) => {
  let url = `/${artifact}/files`;
  if (!Array.isArray(files)) {
    url += `/${files}`;
  } else {
    const body = new URLSearchParams();
    files.forEach((file) => body!.append("files", `${file}`));
    url += "?" + body
  }
  return ajax<void>(url, {method: "DELETE"});
})

// 文件列表（分页）
$on('file', 'list', (data: FilesEventData) => {
  const url = `/${artifact}/files`;
  const query = new URLSearchParams();
  query.set("page", `${data.page}`)
  query.set("limit", `${data.limit}`)
  if (data.directoryId != null) query.set("directoryId", `${data.directoryId}`);
  if (data.mime != null) query.set("mime", data.mime);
  return ajax<FilesEventResult>(url + "?" + query.toString());
})
