import {
  abortion,
  type AbortionContext,
  coerce,
  isInvalid,
  isValid,
  SparkMD5,
  uniqueId
} from "../shared";
import {report} from "./report";
import {Uploader} from "./uploader";

const registry: Map<TaskId, Reader> = new Map();

export function getReadTasks(): Array<Task> {
  return Array.from(registry.values()).map((t) => ({...t.data}));
}

export async function pauseReadTask(id: TaskId, reason?: any): Promise<void> {
  return registry.get(id)?.pause(reason);
}

export async function resumeReadTask(id: TaskId): Promise<void> {
  return registry.get(id)?.resume();
}

export async function removeReadTask(id: TaskId): Promise<void> {
  return registry.get(id)?.destroy();
}

export async function cleanReadTasks(): Promise<void> {
  for (const task of registry.values()) {
    await task.cleanup();
  }
}

/**
 * 读取文件，有效的 5 种任务状态：
 * initial、reading、readend、failed、aborted
 */
export class Reader implements AbortionContext {
  ctrl: AbortController;

  protected constructor(
    readonly file: File,
    readonly data: Task,
  ) {
    this.ctrl = abortion(this);
    this.file = file;
  }

  get id(): TaskId {
    return this.data.id;
  }

  get status(): TaskStatus {
    return this.data.status;
  }

  get signal(): AbortSignal {
    return this.ctrl.signal;
  }

  static create(file: File, dirid: number): void {
    if (!file.type) {
      // 有些文件是没有类型的
      // TODO 需要更加明确的信息来指明是哪一个文件
      void report("app", "error", "未知文件类型");
    } else if (!/^(image|video|audio)\//.test(file.type)) {
      // 仅支持图片、视频和音频这三种类型
      // TODO 需要更加明确的信息来指明是哪一个文件
      void report("app", "error", "不支持的文件类型");
    } else {
      const task = new Reader(file, {
        id: uniqueId(),
        dirid,
        name: file.name,
        path: URL.createObjectURL(file),
        hash: "<hash>",
        mime: file.type,
        size: file.size,
        status: "initial",
      });
      registry.set(task.id, task);
      queueMicrotask(task.start.bind(task));
    }
  }

  /** 暂停任务 */
  pause(reason?: any): void {
    if (isValid(this) && !this.signal.aborted) {
      this.ctrl.abort(reason ?? new Error("已暂停"));
    }
  }

  /** 恢复任务 */
  async resume(): Promise<void> {
    if (isInvalid(this)) {
      if (this.signal.aborted) {
        this.ctrl = abortion(this);
      }
      await this.start();
    }
  }

  /** 清理任务 */
  async cleanup(): Promise<void> {
    if (isInvalid(this)) {
      await this.destroy();
    }
  }

  /** 直接删除 */
  async destroy(): Promise<void> {
    registry.delete(this.id);
    if (!this.signal.aborted) {
      this.ctrl.abort(new Error("删除任务"));
    }
    if (this.status !== "readend") {
      // 如果任务已经完成，文件路径还会被继续使用，
      // 所以只有不在该状态下才能够释放内存。
      URL.revokeObjectURL(this.data.path);
    }
    await report("task", "delete", [this.id]);
  }

  async onabort(_: Event): Promise<void> {
    if (isValid(this)) {
      await this.report("aborted", {
        error: this.signal.reason,
      });
    }
  }

  /** 开始解析文件 */
  private async start(): Promise<void> {
    await this.report("initial");
    if (this.data.hash === "<hash>") {
      await this.report("reading", {progress: 0});
      const spark = new SparkMD5();
      let read = 0;
      try {
        await this.file.stream().pipeTo(
          new WritableStream<Uint8Array>({
            write: async (chunk) => {
              spark.append(chunk);
              read += chunk.length;
              await this.report("reading", {
                progress: Math.min(read / this.file.size, 1),
              });
            },
          }),
          {signal: this.signal},
        );
      } catch (error) {
        if (!this.signal.aborted && isValid(this)) {
          await this.report("failed", {error});
        }
        return;
      }
      this.data.hash = spark.end();
    }
    await this.report("readend", {progress: 1});
    Uploader.create(this.file, this.data, (duplicated: boolean) => {
      registry.delete(this.id);
      if (!duplicated) return;
      return report("app", "error", "任务已经存在了，切勿重复上传哦");
    });
  }

  private async report(
    status: TaskStatus,
    info?: { progress?: number; error?: any },
  ): Promise<void> {
    const error = coerce(info?.error);
    this.data.status = status;
    this.data.progress = info?.progress;
    this.data.error = error != null ? String(error) : undefined;
    await report("task", "sync", {...this.data});
  }
}
