import {
  call,
  coerce,
  createReplier,
  emit,
  isRecvData,
  isSendData,
  on,
  once,
  uniqueId,
} from "../shared";

const ports: Array<MessagePort> = [];
const resolvers: Map<number, VoidFunction> = new Map();

let online = false

// 处理来自客户端的数据
// @ts-ignore
self.onconnect = (evt: MessageEvent): void => {
  const port = evt.ports[0];

  // 客户端卸载事件，无需回复
  once("app:close", () => {
    const index = ports.findIndex((p) => p === port);
    index > -1 && ports.splice(index, 1);
  })

  // 接受客户端消息
  port.onmessage = (evt: MessageEvent): void => {
    if (isRecvData(evt.data)) {
      call(resolvers.get(evt.data.id)); // 忽略来自客户端的结果
    } else if (isSendData(evt.data)) {
      const {scope, action, data} = evt.data
      const reply = createReplier(port, evt.data)
      emit(`${scope}:${action}`, data, reply);
    } else {
      console.warn("[wfs] unknown message event", evt)
    }
  };

  ports.push(port);

  void report('app', 'status', online ? 'online' : 'offline')
}

// WebSocket 事件
on("ws:open", () => {
  online = true
  void report("app", "status", "online")
});

on("ws:close", () => {
  online = false
  void report("app", "status", "offline")
});

on('ws:file:create', (data: CloudFile) => {
  void report('file', 'create', data);
})

on('ws:file:update', (data: CloudFile) => {
  void report('file', 'update', data);
})

export function report(scope: "app", action: "config", config: Partial<ApiConfig>): Promise<void>
export function report(scope: "app", action: "status", data: "online" | "offline"): Promise<void>;
export function report(scope: "app", action: "init", tasks: Array<Task>): Promise<void>;
export function report(scope: "app", action: "error", error: any): Promise<void>;
export function report(scope: "task", action: "sync", task: Task): Promise<void>;
export function report(scope: "task", action: "delete", tasks: number[]): Promise<void>;
export function report(scope: 'file', action: "create", file: CloudFile): Promise<void>
export function report(scope: 'file', action: "update", file: CloudFile): Promise<void>

export function report(scope: string, action: string, data?: any): Promise<void> {
  const id = uniqueId()
  if (scope === "app" && action === "error") {
    data = String(coerce(data));
  }
  return new Promise((resolve) => {
    let sends = 0; // 发送次数
    let count = 0; // 接收次数
    resolvers.set(id, () => {
      if (++count >= sends) {
        resolvers.delete(id);
        resolve();
      }
    });
    queueMicrotask(() => {
      for (const port of ports) {
        sends++
        port.postMessage({id, scope, action, data});
      }
    });
  });
}
