// 参考文档链接
// https://www.zhihu.com/tardis/zm/art/162808604?source_id=1003
// https://www.ruanyifeng.com/blog/2017/05/websocket.html
// https://web.dev/articles/websockets-basics?hl=zh-cn

import {emit} from "../shared";

const HEARTBEAT = "heartbeat";
const CLOSE = "close";

const closeTimeout = 5000; // 等待关闭超时
const reconnectInterval = 3000; // 重连间隔
const heartbeatInterval = 60000; // 心跳间隔
const heartbeatTimeout = 50000; // 心跳超时

// WebSocket 对象引用
let socket: WebSocket | undefined;

// 自动重连开关。
//
// 如果客户端网络不可用，我们就只能等到网络可用时
// 才去连接服务器。所以该变量是用来配合网络变化事件，
// 当连接不可用时能否自动重连。
let isReconnect = navigator.onLine;

// 心跳机制
type Heartbeat = {
  boot: VoidFunction;
  next: VoidFunction;
  stop: VoidFunction;
};

let heartbeat: Heartbeat | undefined;

/**
 * 用于监听网络状态编号时的回调函数
 */
export function handleNetworkEvent(status: "on" | "off"): void {
  // 只有在网络可用的情况下才支持自动重连
  isReconnect = status === "on";

  if (status === "on") {
    // 理论上，切换成在线状态时，立即发送一个心跳包，
    // 这样就可以检测连接是否可以；但是当网络可用时，
    // 打开的连接可能还不能立即恢复，所以延迟发送。
    setTimeout(() => heartbeat?.boot(), reconnectInterval);
  } else {
    heartbeat?.stop();
  }
}

export function useSocket(cfg: Partial<ApiConfig>): void {
  if (cfg.apiUrl && cfg.artifact && cfg.accessToken) {
    const uri = new URL(cfg.apiUrl);
    uri.protocol = uri.protocol.replace(/^http/, "ws");
    uri.pathname = uri.pathname.replace(/\/+$/, "") + `/${cfg.artifact}/notifies`;
    openSocket(uri.toString(), cfg.accessToken);
  } else if (socket != null) {
    closeSocket(socket);
  }
}

function createHeartbeat(ws: WebSocket): Heartbeat {
  let timeoutTimer: ReturnType<typeof setTimeout> | undefined; // 心跳回复超时器
  let heartbeatTimer: ReturnType<typeof setTimeout> | undefined; // 心跳定时器

  const boot = () => {
    // 引用不存在或被重置
    if (socket !== ws) {
      closeSocket(ws);
      return;
    }
    // 如果还存在定时器，就说明没有收到心跳回复，
    // 我们就可以直接关闭连接。
    if (timeoutTimer != null) {
      closeSocket(ws);
      socket = undefined;
      return;
    }
    // 只有在连接可用时才能发送心跳包
    if (ws.readyState === ws.OPEN) {
      // 发送心跳包
      ws.send(HEARTBEAT);
      // 心跳回复超时检测
      timeoutTimer = setTimeout(() => {
        timeoutTimer = undefined;
        closeSocket(ws);
      }, heartbeatTimeout);
    }
  };

  const next = () => {
    // 引用不存在或被重置
    if (socket !== ws) {
      closeSocket(ws);
      return;
    }
    // 取消心跳回复超时事件
    if (timeoutTimer) {
      clearTimeout(timeoutTimer);
      timeoutTimer = undefined;
    }
    // 下次心跳
    heartbeatTimer = setTimeout(() => {
      heartbeatTimer = undefined;
      boot();
    }, heartbeatInterval);
  };

  const stop = () => {
    // 取消心跳回复超时事件
    if (timeoutTimer) {
      clearTimeout(timeoutTimer);
      timeoutTimer = undefined;
    }
    // 取消心跳事件
    if (heartbeatTimer) {
      clearTimeout(heartbeatTimer);
      heartbeatTimer = undefined;
    }
  };

  return {boot, next, stop};
}

// 关闭 WebSocket 连接
function closeSocket(ws: WebSocket, force = false): void {
  console.log("closeSocket")
  switch (ws.readyState) {
    case WebSocket.CONNECTING:
      ws.close();
      return;
    case WebSocket.CLOSING:
    case WebSocket.CLOSED:
      return;
    case WebSocket.OPEN:
      if (force) {
        ws.close();
        return;
      }
      // 我们尝试通知服务器关闭，
      // 如果超过指定时限就强制关闭。
      ws.send(CLOSE);
      setTimeout(() => {
        closeSocket(ws, true);
      }, closeTimeout);
  }
}

interface NotifyData {
  scope: string;
  action: string;
  data: Record<string, any> | Array<any>;
}

function isNotifyData(v: any): v is NotifyData {
  return (
    typeof v === "object" &&
    !Array.isArray(v) &&
    v != null &&
    "scope" in v &&
    "action" in v &&
    "data" in v &&
    typeof v.scope === "string" &&
    typeof v.action === "string" &&
    typeof v.data === "object" &&
    v.data != null
  );
}

// 打开 WebSocket 连接
function openSocket(url: string, protocol: string): void {
  if (socket != null) {
    if (
      socket.url === url &&
      socket.protocol === protocol &&
      socket.readyState < WebSocket.CLOSING
    ) {
      // 如果关键参数一致而且连接是可用的，
      // 我们就不需要重新创建连接。
      return;
    }
    // 关闭之前的连接
    closeSocket(socket);
    socket = undefined;
  }

  // 创建新的连接
  const ws = new WebSocket(url, protocol);
  const hb = createHeartbeat(ws);

  heartbeat = hb;

  // Event: WebSocket opened
  ws.onopen = (): void => {
    if (socket === ws) {
      emit("ws:open"); // 通知连接可用
      hb.boot(); // 启动心跳检测
    } else {
      // 运行到这里，说明在 ws 可用之前就被弃用了，
      // 我们就发送关闭事件通知服务器关闭这个连接。
      closeSocket(ws);
    }
  };

  // Event: WebSocket message received
  ws.onmessage = (evt: MessageEvent): void => {
    // 假定我们只接受字符串消息，服务器不会发送二进制数据
    const lines = (evt.data as string).split("\n");
    if (socket !== ws) {
      // 代码运行到这里，就说明 ws 被弃用了，此刻若收到关闭命令，
      // 我们就主动关闭客户端，相反则通知服务器关闭连接。
      if (lines.includes(CLOSE)) {
        ws.close();
      } else {
        closeSocket(ws);
      }
      return;
    }
    // 处理消息
    for (const line of lines) {
      if (line === CLOSE) {
        // 当我们收到的消息里面包含了关闭命令时，
        // 我们就主动关闭客户端并销毁连接引用。
        // 同时抛弃后续数据
        ws.close();
        socket = undefined;
        return;
      }
      if (line === HEARTBEAT) {
        hb.next();
        continue;
      }
      try {
        const item = JSON.parse(line);
        if (isNotifyData(item)) {
          // 派发事件
          const {scope, action, data} = item;
          emit(`ws:${scope}:${action}`, data);
        } else {
          emit("ws:message", item);
        }
      } catch {
        emit("ws:message", line);
      }
    }
  };

  // Event: WebSocket error
  ws.onerror = (error) => {
    console.error("WebSocket error:", error);
    socket === ws && emit("ws:error", error);
    closeSocket(ws); // Close the socket if an error occurs
  };

  // Event: WebSocket closed
  ws.onclose = () => {
    console.log("WebSocket connection is closed.");
    if (socket === ws) {
      emit("ws:close");
    }
    socket = undefined
    // 停止心跳
    hb.stop();
    // 断线后自动重连
    if (isReconnect) {
      console.log("Reconnecting...");
      setTimeout(function () {
        openSocket(url, protocol);
      }, reconnectInterval);
    }
  };

  socket = ws;
}
