// import {QiniuNetworkError, QiniuRequestError, upload} from './qiniu-js';
// import {type UploadProgress} from "./qiniu-js/upload";
import {
  QiniuNetworkError,
  QiniuRequestError,
  upload,
  type UploadProgress,
} from './qiniu'
import {
  abortion,
  type AbortionContext,
  coerce,
  isInvalid,
  isSameTask,
  isValid,
  promisify,
} from "../shared";
import type {UploadArgs} from './api'
import {
  completeUploadApi,
  initiateUploadApi,
  simulateQiniuCallback
} from './api'
import {report} from "./report";

const registry: Map<FileHash, Uploader> = new Map();

export function getUploadTasks(): Array<Task> {
  return Array.from(registry.values()).reduce<Task[]>(
    (l, u) => l.concat(...u.tasks.map((t) => ({...t}))),
    [],
  );
}

export async function pauseUploadTask(
  hash: FileHash,
  id: TaskId,
): Promise<void> {
  await registry.get(hash)?.pause(id);
}

export async function resumeUploadTask(
  hash: FileHash,
  id: TaskId,
): Promise<void> {
  await registry.get(hash)?.resume(id);
}

export async function removeUploadTask(
  hash: FileHash,
  id: TaskId,
): Promise<void> {
  await registry.get(hash)?.destroy(id);
}

export async function cleanUploadTasks(): Promise<void> {
  for (const tu of registry.values()) {
    await tu.cleanup();
  }
}

/**
 * 后端任务状态
 */
type JobStatus =
  | "initial" // 初始化状态
  | "pending" // 等待上传
  | "uploading" // 正在上传
  | "uploaded" // 上传完成
  | "errored" // 上传失败
  | "aborted"; // 终止上传

const statuses: Array<JobStatus> = [
  "initial", // 初始化状态
  "pending", // 等待上传
  "uploading", // 正在上传
  "uploaded", // 上传完成
  "errored", // 上传失败
  "aborted", // 终止上传
];

export class Uploader implements AbortionContext {
  ctrl: AbortController;

  skipUpload?: boolean;
  uploadToken?: string;
  key?: string;

  status: JobStatus;

  protected constructor(
    readonly file: File,
    readonly hash: FileHash,
    readonly tasks: Array<Task>,
  ) {
    this.ctrl = abortion(this);
    this.status = "initial";
  }

  get signal(): AbortSignal {
    return this.ctrl.signal;
  }

  get isUploaded(): boolean {
    return this.status === "uploaded";
  }

  get isValid(): boolean {
    return statuses.indexOf(this.status) < statuses.indexOf("uploaded");
  }

  get isInvalid(): boolean {
    return statuses.indexOf(this.status) > statuses.indexOf("uploaded");
  }

  static create(
    file: File,
    task: Task,
    callback: (duplicated: boolean) => void,
  ): void {
    let job = registry.get(task.hash);
    if (job == null) {
      job = new Uploader(file, task.hash, [task]);
      registry.set(job.hash, job);
      callback(false);
      queueMicrotask(job.start.bind(job));
      return;
    }
    const exists = job.tasks.some((t) => isSameTask(t, task));
    if (exists) {
      callback(true); // 任务重复
      return;
    }
    job.tasks.push(task);
    callback(false);
    if (!job.isValid) {
      void job.resume(task.id);
    }
  }

  async pause(taskId: TaskId): Promise<void> {
    const task = this.tasks.find((t) => t.id === taskId);
    if (!task) {
      return;
    }
    // 只能中止状态正常的任务
    if (isValid(task)) {
      task.status = "aborted";
      task.error = "已暂停";
      await report("task", "sync", {...task});
    }
    // 如果没有正常的任务，就中止文件上传
    if (this.isValid && !this.signal.aborted && !this.tasks.some(isValid)) {
      this.ctrl.abort();
    }
  }

  // 恢复上传任务
  async resume(taskId: TaskId): Promise<void> {
    // 获取前端任务
    const task = this.tasks.find((t) => t.id === taskId);
    if (task == null) {
      return;
    }
    // 恢复任务
    if (isInvalid(task)) {
      task.status = "readend";
      task.error = undefined;
      task.progress = 1;
      await report("task", "sync", {...task});
    }
    // 确保中断器可用
    if (this.signal.aborted) {
      this.ctrl = abortion(this);
    }
    // 启动任务
    if (this.isInvalid) {
      await this.start();
    }
  }

  // 移除任务
  async destroy(taskId: TaskId): Promise<void> {
    const index = this.tasks.findIndex((t) => t.id === taskId);
    if (index > -1) {
      const task = this.tasks.splice(index, 1)[0];
      URL.revokeObjectURL(task.path); // 释放内存
      await report("task", "delete", [taskId]);
    }
    // 如果没有正常的任务，就中止文件上传
    if (!this.signal.aborted && !this.tasks.some(isValid)) {
      this.ctrl.abort();
    }
    // 若没有前端任务，我们就删除文件上传任务
    if (!this.tasks.length) {
      registry.delete(this.hash);
    }
  }

  async cleanup() {
    let hasValidTask = false;
    const deletes: Array<TaskId> = [];
    for (let i = this.tasks.length - 1; i >= 0; i--) {
      const task = this.tasks[i];
      if (!isValid(task)) {
        this.tasks.splice(i, 1); // 删除任务
        URL.revokeObjectURL(task.path); // 释放内存
        deletes.push(task.id);
      } else {
        hasValidTask = true;
      }
    }
    // 如果没有正常的前端任务，就中止后端任务
    if (!this.signal.aborted && !hasValidTask && !this.isInvalid) {
      this.ctrl.abort();
    }
    // 如果所有的前端任务都被删除了，就同时把后端任务也删除掉。
    if (!this.tasks.length) {
      registry.delete(this.hash);
    }
    // 广播：通知前端删除任务。
    await report("task", "delete", deletes);
  }

  async onabort(_?: Event): Promise<void> {
    await this.report("aborted", {
      status: "aborted",
      error: this.signal.reason || "已暂停",
    });
  }

  async report(
    status: JobStatus,
    info: { error?: any; progress?: number; status: TaskStatus },
  ): Promise<void> {
    this.status = status;
    let error = coerce(info.error);
    if (error != null) error = String(error);
    for (const task of this.tasks) {
      if (isValid(task)) {
        task.status = info.status;
        task.error = error;
        task.progress = info.progress;
        await report("task", "sync", task);
      }
    }
  }

  private async start(): Promise<void> {
    await this.report("initial", {status: "queuing"});
    await this.getUploadOptions();
    await this.startUpload();
    await this.completeUpload();
  }

  private async getUploadOptions(): Promise<void> {
    if (this.skipUpload || this.uploadToken || this.signal.aborted) {
      return;
    }
    try {
      const result = await initiateUploadApi(
        this.file.type,
        this.hash,
        this.signal,
      );
      this.skipUpload = result.skip;
      this.uploadToken = (result as UploadArgs).token;
      this.key = (result as UploadArgs).key;
      this.signal.throwIfAborted();
    } catch (error) {
      if (!this.signal.aborted) {
        await this.report("errored", {
          status: "failed",
          error,
        });
      }
    }
  }

  private async startUpload(): Promise<void> {
    // 任务被中止或者状态错误
    if (this.signal.aborted || this.isUploaded) {
      return;
    }
    // 跳过文件上传
    if (this.skipUpload) {
      await this.report("uploaded", {
        status: "uploaded",
        progress: 1,
      });
      return;
    }
    // 没有状态正常的任务，就不需要上传
    if (!this.tasks.some(isValid)) {
      return;
    }
    const {resolve, promise} = promisify<void>();
    const observable = upload(
      this.file,
      this.key,
      this.uploadToken!,
      {
        fname: this.file.name, // 文件原文件名
        customVars: {"x:md5": this.hash}, // 自定义变量
        mimeType: this.file.type, // 文件类型设置
      },
      {
        useCdnDomain: false,
        checkByServer: true,
        checkByMD5: true,
      },
    );
    const subscription = observable.subscribe({
      next: (res: UploadProgress) => {
        this.report("uploading", {
          status: "uploading",
          progress: res.total.percent / 100,
        });
      },
      error: async (error: any) => {
        if (this.signal.aborted) {
          resolve();
          return;
        }
        // 如果文件已经存在，则模拟七牛云回调，
        // 这样就可以保证文件被正确上传。
        if (
          error instanceof QiniuRequestError &&
          error.data?.error === "file exists"
        ) {
          try {
            await simulateQiniuCallback(
              {
                md5: this.hash,
                name: this.file.name,
                mime: this.file.type,
                size: this.file.size,
                key: this.key!,
              },
              this.signal,
            );
            this.signal.throwIfAborted();
            await this.report("uploaded", {
              status: "uploaded",
              progress: 1,
            });
            resolve();
            return;
          } catch (err) {
            error = err;
          }
        }
        if (!this.signal.aborted) {
          if (error instanceof QiniuNetworkError) {
            await this.report("errored", {
              status: "failed",
              error: "网络异常",
            });
          } else {
            await this.report("errored", {
              status: "failed",
              error
            });
          }
        }
        resolve();
      },
      complete: async (_: any) => {
        // 参数 _ 是七牛云返回我们服务器返回的结果
        await this.report("uploaded", {
          status: "uploaded",
          progress: 1,
        });
        resolve();
      },
    });
    const dispose = (): void => {
      if (!subscription.closed) {
        subscription.unsubscribe();
      }
    };
    // 当信号被中止时取消文件上传
    this.signal.addEventListener("abort", dispose);
    try {
      await promise;
    } finally {
      this.signal.removeEventListener("abort", dispose);
    }
  }

  private async completeUpload(): Promise<void> {
    if (this.signal.aborted || !this.isUploaded) {
      return;
    }
    try {
      const fileNames: Array<string> = [];
      const directories: Array<number> = [];
      // const removes: Array<TaskId> = [];
      for (const task of this.tasks) {
        fileNames.push(task.name);
        directories.push(task.dirid);
        // removes.push(task.id);
      }
      await completeUploadApi(this.hash, fileNames, directories, this.signal);
      await this.report("uploaded", {status: "completed"});
    } catch (error) {
      if (!this.signal.aborted) {
        await this.report("errored", {status: "failed", error});
      }
      return;
    }
    await this.report("uploaded", {
      status: "completed",
      progress: 1,
    });
  }
}
