import {
  computed,
  ComputedRef,
  getCurrentInstance,
  getCurrentScope,
  inject,
  InjectionKey,
  onScopeDispose,
  provide,
  Ref,
  ref,
  watchEffect
} from 'vue'

export interface UseTheme {
  use: Ref<'dark' | 'light' | 'system'>
  dark: ComputedRef<boolean>
  light: ComputedRef<boolean>
}

const key: InjectionKey<UseTheme | undefined> = Symbol("useTheme")

export function useTheme(): UseTheme {
  let instance = inject(key, undefined)
  if (instance != null) {
    return instance
  }

  const use = ref<'dark' | 'light' | 'system'>('light')
  const matches = ref(false)
  const dark = computed(() => {
    if (use.value === 'dark') return true
    if (use.value === 'light') return false
    return matches.value
  })
  const light = computed(() => {
    if (use.value === 'light') return true
    if (use.value === 'dark') return false
    return !matches.value
  })

  const handler = (event: MediaQueryListEvent) => {
    matches.value = event.matches
  }

  let mediaQuery: MediaQueryList | undefined

  const cleanup = () => {
    if (mediaQuery) {
      if ('removeEventListener' in mediaQuery) {
        mediaQuery.removeEventListener('change', handler)
      } else {
        // @ts-expect-error deprecated API
        mediaQuery.removeListener(handler)
      }
    }
  }

  const stopWatch = watchEffect(() => {
    if ('matchMedia' in window && typeof window.matchMedia === 'function') {
      cleanup()

      mediaQuery = window.matchMedia('(prefers-color-scheme: dark)')

      if ('addEventListener' in mediaQuery) {
        mediaQuery.addEventListener('change', handler)
      } else {
        // @ts-expect-error deprecated API
        mediaQuery.addListener(handler)
      }

      matches.value = mediaQuery.matches
    }
  })

  if (getCurrentScope()) {
    onScopeDispose(() => {
      stopWatch()
      cleanup()
      mediaQuery = undefined
    })
  }

  instance = {use, dark, light}

  if (getCurrentInstance()) {
    provide(key, instance)
  }

  return instance
}
