import {store} from './store'
import {installIcons} from './widgets/VIcons'
import UiContainer from './widgets/UiContainer.vue'
import VDialog from './widgets/UiDialog.vue'
import {
  type CSSProperties,
  defineCustomElement,
  getCurrentInstance,
  h,
  nextTick,
  onBeforeMount,
  onMounted,
  onUnmounted,
  type Prop,
  reactive,
  ref,
  type VNode,
  VueElementConstructor,
  watch
} from 'vue'
import style from './style.css?inline'

let FileManager: VueElementConstructor

if (!customElements.get("file-manager")) {
  FileManager = defineCustomElement({
    props: {
      select: {
        type: String,
        validator: v => v == null || v === 'single' || v === 'multiple',
      } as Prop<'single' | 'multiple'>,
      mime: {
        type: String,
        validator: v => v == null || v === 'image' || v === 'video' || v === 'all',
      } as Prop<FileMime>,
      apiUrl: {
        type: String,
        required: true,
      },
      artifact: {
        type: String,
        required: true,
      },
      accessToken: {
        type: String,
        required: true,
      },
      dialog: Boolean,
      width: Number,
      height: Number,
      zIndex: Number
    },
    emits: [
      'close',
      'select',
      'open',
      'fail',
    ],
    setup(props, ctx) {
      const shadowRoot = ref<ShadowRoot | null>(null)
      const visible = ref(false)
      const size = reactive<CSSProperties>({})

      watch(() => props.select, value => {
        store.mode = value ? 'select' : 'manage'
        store.multiple = value === 'multiple'
      }, {immediate: true})

      watch(() => props.mime, value => {
        store.mime = value ?? 'all'
        store.lockMime = value != null
      }, {immediate: true})

      const attachDialog = (root: ShadowRoot) => {
        if (!props.dialog) return
        const style = document.createElement('style')
        style.textContent = `:host{position:fixed;inset:0;z-index:${props.zIndex ?? 100}`
        root.insertBefore(style, root.firstChild)
      }

      const handleWindowResize = () => {
        let {width, height} = props
        if (width == null || width < 100) {
          width = window.innerWidth * 0.8
        }
        if (height == null || height < 100) {
          height = window.innerHeight * 0.8
        }
        size.width = `${Math.max(width, 875)}px`
        size.height = `${Math.max(height, 640)}px`
      }

      onBeforeMount(handleWindowResize)

      onMounted(() => {
        const el = getCurrentInstance()?.vnode.el as HTMLElement
        const root = el.parentNode as ShadowRoot
        installIcons(root)
        attachDialog(root)
        shadowRoot.value = root
        void nextTick(() => {
          visible.value = true
        })
        window.addEventListener('resize', handleWindowResize)
      })

      onUnmounted(() => {
        window.removeEventListener('resize', handleWindowResize)
      })

      const createManager = () => h(UiContainer, {
        setup() {
          return {
            apiUrl: props.apiUrl,
            accessToken: props.accessToken,
            artifact: props.artifact,
          }
        },
        onClose() {
          ctx.emit('close')
        },
        onSelect(files: CloudFile[]) {
          ctx.emit('select', files)
        },
        onFail(message: string) {
          ctx.emit('fail', message)
        }
      })

      return () => {
        let child: VNode
        if (props.dialog && shadowRoot.value) {
          child = h(VDialog, {
            barrier: shadowRoot.value as unknown as Element,
            visible: visible.value,
            class: 'shadow-2xl',
            style: size,
            rounded: false,
            outside: 'shake',
            onDismissed() {
              ctx.emit('close')
            },
            onCompleted() {
              ctx.emit('open')
            }
          }, {
            default: createManager,
          })
        } else {
          child = createManager()
        }
        return h('div', {
          ref: shadowRoot,
          class: ['size-full', {'fixed': props.dialog}]
        }, [child])
      }
    },
    styles: [style],
  })
  customElements.define("file-manager", FileManager)
} else {
  FileManager = customElements.get('file-manager') as VueElementConstructor
}

export {
  FileManager
}

interface ShowFileManagerOptions {
  select?: 'single' | 'multiple',
  mime?: FileMime
  apiUrl: string
  accessToken: string
  artifact: string
  zIndex?: number // with dialog
  width?: number // with dialog
  height?: number // with dialog
  onClose?: VoidFunction // with dialog
  onSelect?: (files: CloudFile[]) => void // with dialog
  onOpen?: VoidFunction // with dialog
  onFail?: (error: string) => void
}

export function showFileManager(opts: ShowFileManagerOptions) {
  let elm: Element | null = new FileManager({
    select: opts.select,
    mime: opts.mime,
    apiUrl: opts.apiUrl,
    accessToken: opts.accessToken,
    artifact: opts.artifact,
    dialog: true,
    zIndex: opts.zIndex,
    width: opts.width,
    height: opts.height,
  })

  const dismiss = () => {
    if (elm != null) {
      elm.remove()
      elm = null
    }
  }

  elm.addEventListener('close', () => {
    opts.onClose?.()
    dismiss()
  })

  elm.addEventListener('select', evt => {
    opts.onSelect?.((evt as CustomEvent).detail as CloudFile[])
    dismiss()
  })

  elm.addEventListener('open', () => {
    opts.onOpen?.()
  })

  elm.addEventListener('fail', evt => {
    opts.onFail?.((evt as CustomEvent).detail)
  })

  document.body.appendChild(elm)

  return dismiss
}
