import {dispatch} from '../worker.ts'
import {coerce, off, on, sleep} from '../../shared'
import {getCurrentScope, nextTick, onScopeDispose, reactive, watch} from 'vue'
import {store} from './store'

export interface Filer {
  loading: boolean
  limit: number
  page: number
  total: number
  error?: string
  dirId: number
  view:
    | "error"
    | "files"
    | "empty"
    | "loader"
}

export function useFiler(directoryId: number): Filer {
  const filer = reactive<Filer>({
    loading: false,
    limit: 30,
    page: 0,
    total: 0,
    dirId: directoryId,
    error: undefined,
    view: "loader",
  })

  const jump = async (to: number) => {
    if (filer.total < 1 || filer.view === "error") {
      filer.view = "loader"
    }

    filer.loading = true
    filer.page = to
    filer.error = undefined

    const mime = store.mime === "all" ? undefined : store.mime
    const {limit, page} = filer
    try {
      const res = await dispatch('file', "list", {
        directoryId,
        page,
        limit,
        mime
      })
      if (filer.page === res.page && filer.limit === res.limit) {
        filer.loading = false
        filer.total = res.total
        filer.view = res.items?.length ? 'files' : 'empty'
        store.files[directoryId] = res.items ?? [] // 这里会不会是响应式的
      }
    } catch (error) {
      if (filer.page === page && filer.limit === limit) {
        filer.view = "error"
        filer.error = String(coerce(error))
        filer.loading = false
      }
    }
  }

  const refresh = async (): Promise<void> => {
    filer.loading = true
    if (filer.page == 0) {
      filer.page = 1
    }
    await sleep(300)
    await jump(filer.page)
  }

  const reload = async (): Promise<void> => {
    filer.loading = true
    filer.view = "loader"
    await sleep(300)
    await jump(filer.page)
  }

  const reset = async (): Promise<void> => {
    filer.page = 0
    filer.total = 0
    delete store.files[directoryId]
    await jump(1)
  }

  watch([
    () => filer.limit,
    () => store.mime,
  ], reset)

  if (getCurrentScope()) {
    on(`files:${directoryId}:refresh`, refresh)
    on(`files:${directoryId}:reload`, reload)
    on(`files:${directoryId}:reset`, reset)
    on(`files:${directoryId}:jump`, jump)
    on("mime:change", reset)

    void nextTick(reset)

    onScopeDispose(() => {
      off(`files:${directoryId}:refresh`, refresh)
      off(`files:${directoryId}:reload`, reload)
      off(`files:${directoryId}:reset`, reset)
      off(`files:${directoryId}:jump`, jump)
      off("mime:change", reset)
      delete store.files[directoryId]
    })
  }

  return filer
}
