import {toast} from '../widgets/UiToastController'
import {dispatch} from '../worker'
import {store} from './store'

export function getDirectoryEntry(id: number): CloudDirectory | undefined {
  if (id === 0) return store.rootDirectory;
  return store.directories.get(id);
}

export function hasChildDirectoryEntries(id: number) {
  return id > 0 && Array.from(store.directories.values()).some((d) => d.pid === id);
}

export async function refreshDirectories() {
  const dirs = await dispatch("dir", "all");
  const set = (dirs: CloudDirectory[] | undefined) => {
    if (dirs == null) return;
    for (const {children, ...dir} of dirs) {
      store.directories.set(dir.id, dir);
      set(children);
    }
  };
  store.directories.clear()
  set(dirs);
  if (store.activeDirectoryId > 0 && !store.directories.has(store.activeDirectoryId)) {
    store.activeDirectoryId = 0;
  }
}

export function reload(): Promise<void> {
  return new Promise<void>((resolve) => {
    store.activeDirectoryId = 0;
    store.files = {};
    store.directories.clear();
    queueMicrotask(() => {
      refreshDirectories()
        .catch(err => toast("error", err))
        .then(resolve)
    });
  });
}

export async function uploadFiles(dirid: number, files: FileList | File[] | File): Promise<void> {
  if (files instanceof File) {
    await dispatch('task', 'create', {dirid, file: files})
  } else {
    for (const file of files) {
      await dispatch('task', "create", {dirid, file})
    }
  }
}

export function deleteFile(file: CloudFile): void {
  const list = store.files[file.directoryId]
  const index = list?.findIndex(f => f.id === file.id)
  if (list && index != null && index > -1) {
    list.splice(index, 1)
  }
  removeSelected(file.id)
}

export function updateFile(file: CloudFile): void {
  const list = store.files[file.directoryId]
  const index = list?.findIndex(f => f.id === file.id)
  if (list && index != null && index > -1) {
    list.splice(index, 1, file)
  }
}

export function useFile(list: CloudFile | Array<CloudFile>) {
  if (!Array.isArray(list)) {
    list = [list]
  }
  for (let file of list) {
    console.log(file.id, file.object.status, file.name)
    const files = store.files[file.directoryId] ?? []
    const index = files.findIndex(f => f.id === file.id)
    if (index === -1) {
      files.push(file)
    } else if (files[index].object.status < file.object.status) {
      files.splice(index, 1, file)
    }
    store.files[file.directoryId] = files
  }
}

export function toggleSelect(file: CloudFile): void {
  if (store.mode === 'manage' || store.multiple) {
    if (!removeSelected(file.id)) {
      store.selected.push({...file})
    }
  } else if (store.selected.length && store.selected[0].id === file.id) {
    store.selected = []
  } else {
    store.selected = [file]
  }
}

export function removeSelected(id: number): boolean {
  const index = store.selected.findIndex(f => f.id === id)
  if (index > -1) {
    store.selected.splice(index, 1)
    return true
  }
  return false
}

// Feature detection. The API needs to be supported
// and the app not run in an iframe.
const supportsFileSystemAccess =
  "showOpenFilePicker" in window &&
  (() => {
    try {
      return window.self === window.top;
    } catch {
      return false;
    }
  })();

export async function pickFiles(directoryId: number) {
  // If the File System Access API is supported…
  if (supportsFileSystemAccess) {
    try {
      // Show the file picker, optionally allowing multiple files.
      // @ts-ignore
      const handles = await self.showOpenFilePicker({
        multiple: true,
        excludeAcceptAllOption: true,
        types: [
          {
            description: "图片和视频",
            accept: {
              "image/*": [
                ".apng",
                ".avif",
                ".gif",
                ".jpeg", ".jpg", ".jfif", ".pjpeg", ".pjp",
                ".png",
                ".webp",
                ".bmp",
                ".ico", ".cur",
                ".tif", ".tiff",
              ],
              "video/*": [
                ".3gp",
                ".mpeg", ".mpg",
                ".mp4", ".m4v", ".m4p",
                ".ogv", ".ogg",
                ".mov", ".qt",
                ".webm",
                ".avi",
              ],
            }
          }
        ],
      });

      const files = await Promise.all(
        handles.map(async (handle: FileSystemFileHandle) => {
          const file = await handle.getFile();
          // Add the `FileSystemFileHandle` as `.handle`.
          // @ts-ignore
          file.handle = handle;
          return file;
        })
      );

      await uploadFiles(directoryId, files)

    } catch (err) {
      // Fail silently if the user has simply canceled the dialog.
      if ((err as Error).name !== 'AbortError') {
        // console.error(err.name, err.message);
        throw err
      }
    }

    return;
  }

  // Fallback if the File System Access API is not supported.
  return new Promise<void>((resolve) => {
    const input = document.createElement('input');
    input.style.display = 'none';
    input.type = 'file';
    input.accept = "image/*,video/*"
    document.body.append(input);
    input.multiple = true;
    input.addEventListener('change', () => {
      input.remove();
      if (!input.files) {
        return;
      }
      uploadFiles(directoryId, input.files)
      resolve()
    });
    // Show the picker.
    if ('showPicker' in HTMLInputElement.prototype) {
      input.showPicker();
    } else {
      input.click();
    }
  });
}
