import Hls from "hls.js"
import Viewer from 'viewerjs'
import viewerStyle from "viewerjs/dist/viewer.css?inline"

const viewerStyleId = "wfs-viewer-style"

let video: HTMLVideoElement | undefined

interface PreviewOptions {
  path: string
  mime: string
  thumb: string
  name: string
}

// https://juejin.cn/post/7146947008015089672
const startVideoPreview = (path: string, img: HTMLImageElement): VoidFunction => {
  video = document.createElement('video');
  video.controls = true
  video.className = img.className
  video.style.cssText = img.style.cssText
    .replace("relative", "absolute")
    .replace("margin-left", "left")
    .replace("margin-top", "top")

  if (Hls.isSupported()) {
    const hls = new Hls();
    hls.loadSource(path);
    hls.attachMedia(video);
    hls.on(Hls.Events.MANIFEST_PARSED, function () {
      void video?.play();
    });
  } else if (video.canPlayType('application/vnd.apple.mpegurl')) {
    video.src = 'https://video-dev.github.io/streams/x36xhzz/x36xhzz.m3u8';
    video.addEventListener('loadedmetadata', function () {
      img.style.opacity = "0"
      void video?.play();
    });
  } else {
    alert("浏览器不支持该视频")
  }

  img.parentNode!.appendChild(video)

  const onResize = () => {
    if (video != null) {
      video.className = img.className
      video.style.cssText = img.style.cssText
        .replace("relative", "absolute")
        .replace("margin-left", "left")
        .replace("margin-top", "top")
    }
  }

  onResize()

  window.addEventListener("resize", onResize)

  return () => {
    video?.pause()
    video?.remove()
    video = undefined
    window.removeEventListener("resize", onResize)
  }
}

export function showPreview(opts: PreviewOptions): void {
  if (!document.querySelector(`#${viewerStyleId}`)) {
    const style = document.createElement("style")
    style.innerHTML = viewerStyle
    style.id = viewerStyleId
    document.head.appendChild(style)
  }

  const isVideo = opts.mime.startsWith("video/")
  const img = document.createElement("img")
  img.src = isVideo ? opts.thumb : opts.path
  let dispose: VoidFunction | undefined
  const viewer = new Viewer(img, {
    title: [4, (_: any, imageData: any) => `${opts.name} (${imageData.naturalWidth} × ${imageData.naturalHeight})`],
    navbar: false,
    zoomable: !isVideo,
    movable: !isVideo,
    backdrop: true,
    slideOnTouch: !isVideo,
    keyboard: !isVideo,
    toolbar: isVideo ? {} : {
      flipHorizontal: true,
      flipVertical: true,
      oneToOne: true,
      reset: true,
      rotateLeft: true,
      rotateRight: true,
      zoomIn: true,
      zoomOut: true,
    },
    viewed(event: CustomEvent) {
      const el = event.detail.image as HTMLImageElement
      isVideo && (dispose = startVideoPreview(opts.path, el))
    },
    hide() {
      dispose?.()
    },
  })
  viewer.show()
}
