/** 事件监听器函数签名 */
type Listener = (...args: any[]) => void;

const events: Record<string, Array<[Listener, boolean]>> = {};

/**
 * 注销事件或取消事件的监听器
 * @param {string} type 事件类型
 * @param {Listener} [listener] 被取消的事件监听器，可选，缺省事会删除指定的事件
 */
export function off(type: string, listener?: Listener): void {
  if (listener == null) {
    delete events[type];
  } else if (Reflect.has(events, type)) {
    const index = events[type].findIndex((x) => x[0] === listener);
    if (index > -1) {
      events[type].splice(index, 1);
    }
  }
}

/**
 * 绑定事件
 * @param {string} type 事件类型
 * @param {Listener} listener 事件监听器
 */
export function on(type: string, listener: Listener): void {
  const listeners = events[type] ?? [];
  listeners[listeners.length] = [listener, false];
  events[type] = listeners;
}

/**
 * 绑定一次性事件
 * @param {string} type 事件类型
 * @param {Listener} listener 事件监听器
 */
export function once(type: string, listener: Listener): void {
  const listeners = events[type] ?? [];
  listeners[listeners.length] = [listener, true];
  events[type] = listeners;
}

/**
 * 触发事件
 *
 * @param {string} type 事件类型
 * @param {Array<*>} args 事件参数列表
 *
 * @returns {void}
 */
export function emit(type: string, ...args: any[]): void {
  if (events[type]) {
    for (let i = events[type].length - 1; i >= 0; i--) {
      const [fn, once] = events[type][i];
      once && events[type].splice(i, 1);
      fn(...args);
    }
  }
}
