import { off, on } from './emitter'
import { coerce } from './utils'

export type Handler<T, R> = (data: T) => Awaitable<R>
export type Replier<R> = (error: any, data?: R) => void

/** 发送给前后端的数据 */
export interface SendData {
  /** 事件数据唯一标识 */
  id: number;
  /** 事件作用域 */
  scope: string;
  /** 事件操作 */
  action: string;
  /** 事件操作参数 */
  data: any;
}

/** 前后端回复的数据 */
export interface RecvData {
  /** 事件数据唯一标识 */
  id: number;
  /** 回复的数据 */
  data?: any
  /** 错误信息 */
  error?: string
}

export function wrap<T, R>(handler: Handler<T, R>) {
  return async (data: T, reply: Replier<R>) => {
    try {
      const res = await handler(data)
      reply(null, res)
    } catch (error) {
      reply(error)
    }
  }
}

export function $on<T, R>(scope: string, action: string, handler: Handler<T, R>): VoidFunction {
  const listener = wrap(handler)
  on(`${scope}:${action}`, listener)
  return () => off(`${scope}:${action}`, listener)
}

/**
 * 发送给前后端的数据
 * @param {any} v
 * @returns {boolean}
 */
export function isSendData(v: any): v is SendData {
  return (
    typeof v === "object" &&
    !Array.isArray(v) &&
    v != null &&
    "id" in v &&
    "scope" in v &&
    "action" in v &&
    "data" in v &&
    Reflect.ownKeys(v).length === 4
  );
}

/**
 * 前后端回复的数据
 * @param {any} v
 * @returns {boolean}
 */
export function isRecvData(v: any): v is RecvData {
  return (
    typeof v === "object" &&
    !Array.isArray(v) &&
    v != null &&
    "id" in v &&
    typeof v.id === 'number' &&
    Reflect.ownKeys(v).length === ("data" in v || 'error' in v ? 2 : 1)
  );
}

/**
 * 创建回复数据
 * @param {MessagePort} port
 * @param {SendData} send
 * @returns {Replier<any>}
 */
export function createReplier(port: MessagePort, send: SendData): Replier<any> {
  return (error: any, data?: any): void => {
    const recv: RecvData = Object.create(null);
    recv.id = send.id;
    if (error != null) {
      recv.error = String(coerce(error))
    } else if (data != null) {
      recv.data = data
    }
    port.postMessage(recv)
  }
}
