/// <reference types="vite/client" />

type Awaitable<T> = T | Promise<T>;
type Arrayable<T> = T | Array<T>

/** 目录信息 */
interface CloudDirectory {
  id: number;
  pid: number;
  title: string;
  sort: number;
  createdAt: string;
  children?: CloudDirectory[];
}

/** 文件基本信息 */
interface CloudFile {
  id: number;
  appId: number;
  objectId: number;
  directoryId: number;
  name: string;
  mime: string;
  object: CloudObject;
  createdBy: number;
  createdAt: string;
  updatedAt: string;
}

/** 文件对象信息 */
interface CloudObject {
  id: number;
  name: string;
  hash: string;
  size: number;
  source: string;
  path: string;
  thumb: string;
  mime: string;
  width: number;
  height: number;
  duration: number;
  status: number;
  error?: string;
  createdAt: string;
  updatedAt: string;
}

type FileMime =
  | 'image' // 图片
  | 'video' // 视频
  | 'all' // 全部

type FileMode =
  | "manage" // 管理模式
  | "select" // 选择模式

/**
 * 表示任务的编号
 *
 * 由于 ReadTask 和 UploadTask 存在于用户设备上，
 * 我们可以把它们看做是本地的任务，规定其编号为正整数；
 *
 * 而任务 CloudTask 表示云端任务，规定其编号为负整数。
 */
type TaskId = number;

/**  表示文件内容的 MD5 值 */
type FileHash = string;

/** 任务状态有效值 */
type TaskStatus =
  | "initial" // 任务刚刚创建
  | "reading" // 正在读取文件内容
  | "readend" // 文件内容读取完成
  | "queuing" // 已经将任务添加到上传队列中
  | "uploading" // 正在上传任务文件
  | "uploaded" // 任务文件上传完成
  | "completed" // 任务完成
  | "failed" // 任务上传失败，比如网络原因
  | "aborted"; // 上传任务被用户中止

/**
 * 任务属性
 *
 * 通过 dirid + name 来确保任务的唯一性；
 *
 * 初始任务时属性 hash 的值与 id 保持一致，
 * 只有当文件读取完成时才会使用正确值填充。
 */
interface Task {
  /** 任务编号 */
  id: TaskId;
  /** 目标目录 */
  dirid: number;
  /** 文件名称 */
  name: string;
  /** 文件 MD5 值 */
  hash: FileHash;
  /** 文件媒体类型 */
  mime: string;
  /** 文件路径 */
  path: string;
  /** 文件大小 */
  size: number;
  /** 相关进度 */
  progress?: number;
  /** 错误信息 */
  error?: string;
  /** 任务状态 */
  status: TaskStatus;
}

/** 发送给前后端的数据 */
interface SendData {
  /** 事件数据唯一标识 */
  id: number;
  /** 事件作用域 */
  scope: string;
  /** 事件操作 */
  action: string;
  /** 事件操作参数 */
  data: any;
}

/** 前后端回复的数据 */
interface RecvData {
  /** 事件数据唯一标识 */
  id: number;
  /** 回复的数据 */
  data?: any;
  /** 错误信息 */
  error?: string;
}

/** 接口配置 */
interface ApiConfig {
  apiUrl: string;
  accessToken: string;
  artifact: string;
}

/** 创建上传任务 */
interface TaskCreateData {
  dirid: number;
  file: File;
}

interface TaskEventData {
  fileHash: FileHash;
  taskId: TaskId;
}

interface NewDirEventData {
  title: string;
  pid?: number;
}

interface RenameDirEventData {
  id: number;
  title: string;
}

interface RenameFileEventData {
  id: number;
  name: string;
}

interface FilesEventData {
  directoryId: number | undefined;
  page: number;
  limit: number;
  mime?: string;
}

interface FilesEventResult {
  total: number;
  page: number;
  limit: number;
  items: CloudFile[] | undefined;
}

// export declare type RunTask<T> = (...args: T[]) => Promise<void>;
// export interface QueueContent<T> {
//   task: T;
//   resolve: () => void;
//   reject: (err?: any) => void;
// }
// export declare class Pool<T> {
//   private runTask;
//   private limit;
//   aborted: boolean;
//   queue: Array<QueueContent<T>>;
//   processing: Array<QueueContent<T>>;
//   constructor(runTask: RunTask<T>, limit: number);
//   enqueue(task: T): Promise<void>;
//   private run;
//   private check;
//   abort(): void;
// }
